#!/usr/bin/env python3
import os
import unittest

import caosdb as db


def get_entity_with_id(eid):
    return db.execute_query("FIND "+str(eid), unique=True)


class CrawlerTest(unittest.TestCase):
    def test_experiment(self):
        ########################
        # # first experiment # #
        ########################
        # TODO saving an empty string as value in a text property leads to a
        # vanishing of the property
        # thus an x is used here. Needs to be fixed.
        exp = db.execute_query(
            "FIND Experiment with date=2019-02-03 and identifier=empty_identifier",
            unique=True)

        # There should be a Project with name TestProject which is referenced
        project_id = exp.get_property("Project").value
        project = get_entity_with_id(project_id)
        assert project.name == "2010_TestProject"
        assert "Project" in [p.name for p in project.get_parents()]

        # There should be a datafile as result attached with path datafile.dat
        datfile_id = exp.get_property("results").value[0]
        datfile = get_entity_with_id(datfile_id)
        self.assertEqual("an example reference to a results file",
                         datfile.description)
        assert os.path.basename(datfile.path) == "datafile.dat"

        # Should have a responsible person
        self.assertIsNotNone(exp.get_property("responsible"))
        person = db.Record(id=exp.get_property("responsible").value[0])
        person.retrieve()
        self.assertEqual("Only", person.get_property("firstname").value)

        # Should have a description
        self.assertIsNotNone(exp.description)

        #########################
        # # second experiment # #
        #########################
        exp = db.execute_query(
            "FIND Experiment with date=2019-02-03 and identifier='something'",
            unique=True)

        # Should be the same project
        assert project_id == exp.get_property("Project").value

        # Should have two data files
        datfile_ids = exp.get_property("results").value
        datafile_names = []

        for did in datfile_ids:
            datfile = get_entity_with_id(did)
            datafile_names.append(os.path.basename(datfile.path))
        assert "usefull.xlsx" in datafile_names
        assert "useless.xlsx" in datafile_names

    def test_analysis(self):
        ######################
        # # first analysis # #
        ######################
        ana = db.execute_query(
            "FIND Analysis with date=2019-02-03 and identifier='empty_identifier'",
            unique=True)

        # There should be a Project with name TestProject which is referenced
        project_id = ana.get_property("Project").value
        project = get_entity_with_id(project_id)
        assert "Project" in [p.name for p in project.get_parents()]

        # There should be a file as result attached with path results.pdf
        datfile_id = ana.get_property("results").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "results.pdf"

        # There should be a file as script attached with path plot.py
        datfile_id = ana.get_property("scripts").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "plot.py"

        # Should have a responsible person
        self.assertIsNotNone(ana.get_property("responsible"))
        person = db.Record(id=ana.get_property("responsible").value[0])
        person.retrieve()
        self.assertEqual("Only", person.get_property("firstname").value)

        # Should have a description
        self.assertIsNotNone(ana.description)

        # Should reference an experiment
        sources = [get_entity_with_id(el) for el in
                   ana.get_property("sources").value]
        count = 0

        for s in sources:
            if len(s.parents) > 0 and s.parents[0].name == "Experiment":
                count += 1
        self.assertEqual(count, 1)
        self.assertEqual(len(sources), 2)

        first_ana = ana

        #######################
        # # second analysis # #
        #######################
        ana = db.execute_query(
            "FIND Analysis with date=2019-02-03 and identifier='something'",
            unique=True)

        # Should be the same project
        assert project_id == ana.get_property("Project").value

        # There should be a file as script attached with path plot.py
        datfile_id = ana.get_property("scripts").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "analyse.py"

        # Should reference an experiment
        sources = [get_entity_with_id(el) for el in
                   ana.get_property("sources").value]
        count = 0

        for s in sources:
            if len(s.parents) > 0 and s.parents[0].name == "Experiment":
                count += 1
        self.assertEqual(count, 1)
        self.assertEqual(len(sources), 1)

        # Should reference the other analysis
        self.assertIsNotNone(ana.get_property("revisionOf"))
        a = db.Record(id=ana.get_property("revisionOf").value[0])
        a.retrieve()
        self.assertEqual(a.id, first_ana.id)

    def test_simulation(self):
        ######################
        # # first simulation # #
        ######################
        sim = db.execute_query(
            "FIND Simulation with date=2019-02-03 and identifier='empty_identifier'",
            unique=True)

        # There should be a Project with name TestProject which is referenced
        project_id = sim.get_property("Project").value
        project = get_entity_with_id(project_id)
        assert "Project" in [p.name for p in project.get_parents()]

        # There should be a file as result attached with path results.pdf
        datfile_id = sim.get_property("results").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "snapshots.dat"

        # There should be a file as script attached with path plot.py
        datfile_id = sim.get_property("scripts").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "sim.py"

        # Should have a responsible person
        self.assertIsNotNone(sim.get_property("responsible"))
        person = db.Record(id=sim.get_property("responsible").value[0])
        person.retrieve()
        self.assertEqual("Only", person.get_property("firstname").value)

        # Should have a description
        self.assertIsNotNone(sim.description)

        #########################
        # # second simulation # #
        #########################
        sim = db.execute_query(
            "FIND Simulation with date=2019-02-03 and identifier='something'",
            unique=True)

        sources = [get_entity_with_id(el) for el in
                   sim.get_property("sources").value]
        count = 0

        for s in sources:
            if len(s.parents) > 0 and s.parents[0].name == "Experiment":
                count += 1
        self.assertEqual(count, 1)
        self.assertEqual(len(sources), 1)

        # Should be the same project
        assert project_id == sim.get_property("Project").value

        # Should have two responsible persons
        self.assertIsNotNone(sim.get_property("responsible"))
        self.assertEqual(len(sim.get_property("responsible").value), 2)

        # TODO enable revisionOf
        # Should have revision
        # osim = get_entity_with_id(sim.get_property("revisionOf").value[0])
        # self.assertEqual(osim.parents[0].name, "Simulation")
        # self.assertNotEqual(osim.id, sim.id)

        # Should have data file
        datfile_id = sim.get_property("results").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "timeseries.npy"

        # Should have two scritps files
        datfile_ids = sim.get_property("scripts").value
        datafile_names = []

        for did in datfile_ids:
            datfile = get_entity_with_id(did)
            datafile_names.append(os.path.basename(datfile.path))
        assert "parameters.p" in datafile_names
        assert "large_sim.py" in datafile_names

    def test_publication(self):
        #########################
        # # first publication # #
        #########################
        pub = db.execute_query("FIND *really_cool_finding", unique=True)

        # There should be a file as result attached with path poster.pdf
        datfile_id = pub.get_property("results").value[0]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "poster.pdf"

        # There should be a file as data attached
        datfile_id = pub.get_property("sources").value[1]
        datfile = get_entity_with_id(datfile_id)
        assert os.path.basename(datfile.path) == "results.pdf"

        # Test type
        self.assertEqual(pub.parents[0].name, "Poster")

        ##########################
        # # second publication # #
        ##########################
        pub = db.execute_query("FIND *paper_on_exciting_stuff ", unique=True)

        # Test type
        self.assertEqual(pub.parents[0].name, "Thesis")
